/*
  written by izumi@ieee.org, Mar 2019.
  gcc -O -Wall -g test.c -mwindows -lws2_32 -lmswsock -ladvapi32 -lwinmm
*/

#define GEISTERCLIENT_C

#define DEBUG 1

#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <stdlib.h>
#include <stdio.h>
#include <w32api/mmsystem.h>
#include <string.h>

#include "geister.h"
#include "geisterclient.h"

static WSADATA wsadata;
#define BUFSZ 256

/* ---------------------------------------------------------------- */
SOCKET ConnectGeisterHost(char *hostaddr, char *hostport){
  SOCKET gsock=INVALID_SOCKET;
  struct addrinfo hints, *arslt=NULL, *aptr=NULL;
  int r=0;
  
  ZeroMemory(&hints,sizeof(hints));
  hints.ai_family  =AF_UNSPEC;
  hints.ai_socktype=SOCK_STREAM;
  hints.ai_protocol=IPPROTO_TCP;

  // initialize winsock
  r=WSAStartup(MAKEWORD(2,2),&wsadata);
  if (r!=0) {
    fprintf(stderr,"WSAStartup failed with error: %d\n",r);
    return INVALID_SOCKET;
  }

  //resolve the server address and port
  r=getaddrinfo(hostaddr,hostport,&hints,&arslt);
  if (r!=0) {
    fprintf(stderr,"getaddrinfo failed with error: %d\n",r);
    WSACleanup();
    return INVALID_SOCKET;
  }

  //attempt to connect
  for(aptr=arslt; aptr!=NULL ;aptr=aptr->ai_next) {
    //create a SOCKET for connecting to server
    gsock=socket(aptr->ai_family,aptr->ai_socktype,aptr->ai_protocol);
    if (gsock==INVALID_SOCKET) {
      fprintf(stderr,"socket failed with error: %d\n", WSAGetLastError());
      WSACleanup();
      return INVALID_SOCKET;
    }
    //connect to server
    r=connect(gsock,aptr->ai_addr,(int)aptr->ai_addrlen);
    if (r!=SOCKET_ERROR) {
      // connected successfully
      freeaddrinfo(arslt);
      return gsock;
    }
    closesocket(gsock);
  }

  freeaddrinfo(arslt);
  
  fprintf(stderr,"unable to connect to server!\n");
  WSACleanup();
  return INVALID_SOCKET;
}

/* ---------------------------------------------------------------- */
int CloseGeisterHost(SOCKET gsock){
  closesocket(gsock);
  WSACleanup();
  return 0;
}

/* ---------------------------------------------------------------- */
void DieGeister(SOCKET gsock,char *msg){
  fprintf(stderr,"%s\n",msg);
  CloseGeisterHost(gsock);
  exit(1);
}

/* ---------------------------------------------------------------- */
int RecvLine(SOCKET gsock, char buf[]){
  static char buf0[BUFSZ];
  static int buf0c=0;
  int n=0,r,i,j;

  for (i=0;i<BUFSZ;i++) buf[i]='\0';

  while (1) {
    r=recv(gsock,buf0+buf0c,BUFSZ-1-buf0c,0);
    if (r==SOCKET_ERROR) DieGeister(gsock,"error in recv(socket)\n");
    buf0c=buf0c+r;
    if (buf0c>=BUFSZ-1) {
      n=BUFSZ-2;
      goto ENDOFLINE;
    }
    for (n=0;n<buf0c;n++) {
      if (buf0[n]=='\n') goto ENDOFLINE;
    }
    fprintf(stderr,"waiting message from host\n");
    Sleep(1000);
  }
 ENDOFLINE:
  n++;
  for (i=0;i<n;i++) buf[i]=buf0[i];
  buf[n]='\0';
  for (i=n,j=0;i<buf0c;i++,j++) buf0[j]=buf0[i];
  buf0c=j;

  return n;
}

/* ---------------------------------------------------------------- */
int SendGhostInfo(SOCKET gsock, struct STATE *state, int player_com) {
  char buf[BUFSZ];
  int r,i;
  int red, blue;
  
  red  = ghostcolorid[player_com][RED];
  blue = ghostcolorid[player_com][BLUE];
  
  //receive message (SET?)
  r=RecvLine(gsock,buf);
  fprintf(stderr,"recv req (%s)\n",buf);
  if (strncmp(buf,"SET?",4)!=0){
    fprintf(stderr,"SendGhostInfo: recv wrong msg (%s) while expecting (SET?)\n",buf);
    DieGeister(gsock,"");
  }
  
  // send ghost setting
  strncpy(buf,"SET:XXXX\r\n",BUFSZ);
  i=4;// first X
  if (state->board[1][4]==red) buf[i++]='A';
  if (state->board[2][4]==red) buf[i++]='B';
  if (state->board[3][4]==red) buf[i++]='C';
  if (state->board[4][4]==red) buf[i++]='D';
  if (state->board[1][5]==red) buf[i++]='E';
  if (state->board[2][5]==red) buf[i++]='F';
  if (state->board[3][5]==red) buf[i++]='G';
  if (state->board[4][5]==red) buf[i++]='H';
  if (state->board[4][1]==red) buf[i++]='A';
  if (state->board[3][1]==red) buf[i++]='B';
  if (state->board[2][1]==red) buf[i++]='C';
  if (state->board[1][1]==red) buf[i++]='D';
  if (state->board[4][0]==red) buf[i++]='E';
  if (state->board[3][0]==red) buf[i++]='F';
  if (state->board[2][0]==red) buf[i++]='G';
  if (state->board[1][0]==red) buf[i++]='H';
  fprintf(stderr,"send ghost setting (%s)\n",buf);
  r=send(gsock,buf,strlen(buf),0);
  if (r==SOCKET_ERROR) DieGeister(gsock,"error in send(socket)\n");

  //receive message (OK?)
  r=RecvLine(gsock,buf);
  fprintf(stderr,"recv ack (%s)\n",buf);
  if (strncmp(buf,"OK",2)!=0) {
    fprintf(stderr,"SendGhostInfo: recv wrong msg (%s) while expecting (OK)\n",buf);
    DieGeister(gsock,"");
  }
  
  return 1;
}

/* ---------------------------------------------------------------- */
int RecvBoardState(SOCKET gsock,struct STATE *state, int player_com){
  char buf[BUFSZ];
  int c,p,i;
  //
  int j;
  for(j=0;j<6;j++){
    for(i=0;i<6;i++){
      state->board[i][j] = EMPTY;
    }
  }
  //
  
  //receive message 
  c=RecvLine(gsock,buf);
  fprintf(stderr,"received board state (%s)\n",buf);

  if (strncmp(buf,"WON",3)==0) return WIN;
  if (strncmp(buf,"LST",3)==0) return LOSE;

  //translate ghost info
  for (p=0;p<2;p++) {
    int ofs;
    ofs = (p==player_com)?4:4+8*3; // offset for reading buf
    state->ghost_num[p][RED ]=4;
    state->ghost_num[p][BLUE]=4;
    for (i=0;i<8;i++) {
      int x,y,c;
      switch (buf[ofs+i*3+2]){
      case 'R':case 'r':c=RED; break;
      case 'B':case 'b':c=BLUE;break;
      default:          c=UNKNOWN;
      }
      //fprintf(stderr,"ghost[%d][%d]=%c%c%c\n",p,i,buf[ofs+i*3+0],buf[ofs+i*3+1],buf[ofs+i*3+2]);
      state->ghost[p][i].c=c;
      x=buf[ofs+i*3+0]-'0';
      y=buf[ofs+i*3+1]-'0';
      //fprintf(stderr,"ghost[%d][%d](x,y,c)=(%d,%d,%d)\n",p,i,x,y,c);
      if (x>=0&&x<=5&&y>=0&&y<=5) {
	// on board
	state->ghost[p][i].x=(player_com==PLAYER0)?x:5-x;
	state->ghost[p][i].y=(player_com==PLAYER0)?y:5-y;
	switch (c) {
	case RED:  state->board[x][y]=ghostcolorid[p][c];break;
	case BLUE: state->board[x][y]=ghostcolorid[p][c];break;
	default:   state->board[x][y]=UNKNOWN;
	}
      } else if (x==8&&y==8) {
	// escaped
	state->ghost[p][i].x=x;
	state->ghost[p][i].y=y;
      } else if (x==9&&y==9) {
	// captured
	state->ghost[p][i].x=x;
	state->ghost[p][i].y=y;
	switch (c) {
	case RED:  state->ghost_num[p][RED ]--; break;
	case BLUE: state->ghost_num[p][BLUE]--; break;
	default: DieGeister(gsock,"captured but unknown");
	}
      } else {
	DieGeister(gsock,"ghost is out of board");
      }
    }
  }

  return GAMECONTINUE;
}

/* ---------------------------------------------------------------- */
int SendAction(SOCKET gsock, struct STATE *state, int player_com, struct ACTION *action){
  char buf[BUFSZ];
  int n,r;

  fprintf(stderr,"SendAction(%d,%d,%d)\n",action->x,action->y,action->dir);

  // find the ghost at the position
  for (n=0;n<8;n++)
    if (state->ghost[player_com][n].x==action->x&&
	state->ghost[player_com][n].y==action->y) break;
  if (n>=8){
    fprintf(stderr,"no ghost on (%d,%d)\n",action->x,action->y);
    DieGeister(gsock,"");
  }

  // send ghost setting
  strncpy(buf,"MOV:X,D\r\n",BUFSZ);
  buf[4]='A'+n;
  switch(player_com){
  case PLAYER0:
    switch(action->dir){
    case NORTH:buf[6]='N';break;
    case SOUTH:buf[6]='S';break;
    case EAST :buf[6]='E';break;
    case WEST :buf[6]='W';break;
    default:DieGeister(gsock,"unknown direction");
    }
    break;
  case PLAYER1:
    switch(action->dir){
    case NORTH:buf[6]='S';break;
    case SOUTH:buf[6]='N';break;
    case EAST :buf[6]='W';break;
    case WEST :buf[6]='E';break;
    default:DieGeister(gsock,"unknown direction");
    }
    break;
  default:DieGeister(gsock,"unknown direction");
  }
  fprintf(stderr,"send action (%s)\n",buf);
  r=send(gsock,buf,strlen(buf),0);
  if (r==SOCKET_ERROR) DieGeister(gsock,"error in send(socket)\n");

  //receive message (OK?)
  r=RecvLine(gsock,buf);
  fprintf(stderr,"recv ack (%s)\n",buf);
  if (strncmp(buf,"OK",2)!=0) {
    fprintf(stderr,"recv wrong msg (%s) while expecting (OK)\n",buf);
    DieGeister(gsock,"");
  }

  return 1;
}

/* ---------------------------------------------------------------- */
